defmodule ClickHouse do
  @moduledoc """
  A ClickHouse client.

  This currently represents an early work in progress.
  """

  @type client :: atom()
  @type statement :: binary()
  @type params :: list()
  @type interface :: module()
  @type format :: module()
  @type option ::
          {:client, client()}
          | {:interface, interface()}
          | {:interface_opts, keyword()}
  @type options :: [option()]
  @type error ::
          ClickHouse.ConnectionError.t()
          | ClickHouse.DatabaseError.t()
          | ClickHouse.FormatError.t()
          | ClickHouse.NetworkError.t()
          | ClickHouse.ParsingError.t()

  use Supervisor

  alias ClickHouse.{Config, Query, Result}

  @formats [
    {ClickHouse.Format.TSV, ["TabSeparated", "TSV"]},
    {ClickHouse.Format.TSVWithNames, ["TabSeparatedWithNames", "TSVWithNames"]},
    {ClickHouse.Format.TSVWithNamesAndTypes,
     ["TabSeparatedWithNamesAndTypes", "TSVWithNamesAndTypes"]}
  ]

  @opts_schema %{
    client: [type: :atom, required: true, default: :default],
    interface: [type: :module, required: true, default: ClickHouse.Interface.HTTP],
    formats: [
      type: {:list, {:tuple, {:module, {:list, :binary}}}},
      required: true,
      default: @formats
    ]
  }

  ################################
  # Public API
  ################################

  @doc """
  Starts a ClickHouse client.

  ## Options

    * `:client` - The name of the client - defaults to `:default`.
    * `:interface` - The interface to use for queries - defaults to `ClickHouse.Interface.HTTP`.
    * `:formats` - A list of formats available for encoding/decoding.

  ## Extra Options

  Any additional options passed will be given to the client interface.
  """
  @spec start_link(options()) :: Supervisor.on_start()
  def start_link(opts \\ []) do
    {opts, extra_opts} = Keyword.split(opts, Map.keys(@opts_schema))
    opts = KeywordValidator.validate!(opts, @opts_schema)
    Supervisor.start_link(__MODULE__, {opts, extra_opts})
  end

  @doc """
  Prepares and executes a query using a ClickHouse client.
  """
  @spec query(client(), statement(), params(), keyword()) ::
          {:ok, ClickHouse.Result.t()} | {:error, error()}
  def query(client, statement, params \\ [], opts \\ []) do
    with {:ok, query} <- prepare(client, statement, params, opts) do
      execute(query)
    end
  end

  @doc """
  Prepares a query for a ClickHouse client.
  """
  @spec prepare(client(), statement(), params(), keyword()) ::
          {:ok, ClickHouse.Query.t()} | {:error, ClickHouse.FormatError.t()}
  def prepare(client, statement, params \\ [], opts \\ []) do
    client
    |> Query.new(statement, params, opts)
    |> Query.encode()
  end

  @doc """
  Executes a query using a ClickHouse client.
  """
  @spec execute(ClickHouse.Query.t()) :: {:ok, ClickHouse.Result.t()} | {:error, error()}
  def execute(%{interface: nil}), do: raise(ArgumentError, "No query interface")

  def execute(query) do
    with {:ok, result} <- query.interface.execute(query) do
      Result.decode(result)
    end
  end

  ################################
  # Supervisor Callbacks
  ################################

  @impl Supervisor
  def init({opts, extra_opts}) do
    init_config(opts)

    children = [
      build_interface(opts, extra_opts)
    ]

    Supervisor.init(children, strategy: :one_for_one)
  end

  ################################
  # Private API
  ################################

  defp init_config(opts) do
    client = Keyword.fetch!(opts, :client)
    Config.new(client, opts)
  end

  defp build_interface(opts, extra_opts) do
    client = Keyword.fetch!(opts, :client)
    interface = Keyword.fetch!(opts, :interface)
    interface_opts = Keyword.put(extra_opts, :client, client)

    {interface, interface_opts}
  end
end
