if Code.ensure_loaded?(NimbleCSV) do
  defmodule ClickHouse.Format.TSV do
    @moduledoc """
    An implementation of the ClickHouse `"TSV"`/`"TabSeparatedValues"` format.
    """

    @behaviour ClickHouse.Format

    alias ClickHouse.{FormatError, Query}

    @parser __MODULE__.Parser

    NimbleCSV.define(@parser, moduledoc: false, separator: "\t", escape: "\"")

    ################################
    # ClickHouse.Format Callbacks
    ################################

    @impl ClickHouse.Format
    @spec decode(ClickHouse.Result.t()) :: {:ok, ClickHouse.Result.t()}
    def decode(result), do: {:ok, do_decode(result)}

    @impl ClickHouse.Format
    @spec encode(ClickHouse.Query.t()) ::
            {:ok, ClickHouse.Query.t()} | {:error, ClickHouse.FormatError.t()}
    def encode(%Query{command: :insert, params_count: count}) when count > 1 do
      error = %FormatError{message: "TSV insert statement can only support a single parameter."}
      {:error, error}
    end

    def encode(%Query{command: :insert, params_count: 0, params: [_ | _]}) do
      error = %FormatError{message: "TSV insert statement with params must include a '$1'."}
      {:error, error}
    end

    def encode(%Query{command: :insert, params_count: 1, params: params} = query) do
      data = do_encode(params)
      encoded = String.replace(query.statement, "$1", data)
      {:ok, Query.put_encoded(query, encoded)}
    end

    def encode(%Query{command: :select} = query) do
      encoded = encode_params(query.statement, 1, query.params)
      {:ok, Query.put_encoded(query, encoded)}
    end

    def encode(query) do
      {:ok, query}
    end

    ################################
    # Private API
    ################################

    defp do_decode(result) do
      data = @parser.parse_string(result.raw, skip_headers: false)
      %{result | data: data}
    end

    defp do_encode(params) do
      params
      |> Enum.map(&encode_row/1)
      |> @parser.dump_to_iodata()
      |> IO.iodata_to_binary()
    end

    defp encode_row(values), do: Enum.map(values, &encode_value/1)

    defp encode_value(nil), do: nil
    defp encode_value(false), do: 0
    defp encode_value(true), do: 1
    defp encode_value(string) when is_binary(string), do: string
    defp encode_value(integer) when is_integer(integer), do: to_string(integer)
    defp encode_value(atom) when is_atom(atom), do: to_string(atom)
    defp encode_value(%Date{} = date), do: to_string(date)

    defp encode_value(%DateTime{} = datetime) do
      datetime
      |> DateTime.truncate(:second)
      |> DateTime.to_iso8601()
      |> String.replace("Z", "")
    end

    defp encode_value(list) when is_list(list) do
      values = Enum.map(list, &encode_array_value/1)
      "[" <> Enum.join(values, ",") <> "]"
    end

    defp encode_array_value(nil), do: nil
    defp encode_array_value(false), do: 0
    defp encode_array_value(true), do: 1
    defp encode_array_value(integer) when is_integer(integer), do: to_string(integer)

    defp encode_array_value(string) when is_binary(string) do
      string |> with_quote()
    end

    defp encode_array_value(atom) when is_atom(atom) do
      atom |> to_string() |> with_quote()
    end

    defp encode_array_value(%Date{} = date) do
      date |> to_string() |> with_quote()
    end

    defp encode_array_value(%DateTime{} = datetime) do
      datetime |> encode_value() |> with_quote()
    end

    defp encode_array_value(list) when is_list(list) do
      values = Enum.map(list, &encode_array_value/1)
      "[" <> Enum.join(values, ",") <> "]"
    end

    defp with_quote(value) do
      "'" <> value <> "'"
    end

    defp encode_params(statement, _count, []), do: statement

    defp encode_params(statement, count, [param | params]) do
      statement
      |> encode_param(count, param)
      |> encode_params(count + 1, params)
    end

    defp encode_param(statement, count, param) do
      param = encode_array_value(param)
      String.replace(statement, "$#{count}", param)
    end
  end
end
