defmodule ClickHouse.Result do
  @moduledoc """
  The results of a ClickHouse query.
  """

  alias ClickHouse.Config

  @type name :: binary()
  @type type :: atom() | nil
  @type columns :: [{name(), type()}, ...] | nil
  @type data :: [] | [list(), ...] | nil
  @type t :: %__MODULE__{
          client: ClickHouse.client(),
          raw: binary(),
          raw_format: binary(),
          format: ClickHouse.format(),
          columns: columns(),
          data: data(),
          compressed: boolean(),
          meta: map()
        }

  @derive {Inspect, only: [:data, :columns]}
  defstruct [:client, :raw, :raw_format, :format, :columns, :meta, :compressed, :data]

  ################################
  # Public API
  ################################

  @spec new(ClickHouse.client(), binary(), binary(), map(), boolean()) :: ClickHouse.Result.t()
  def new(client, raw, raw_format, meta, compressed) do
    %__MODULE__{
      client: client,
      raw: raw,
      raw_format: raw_format,
      meta: meta,
      compressed: compressed
    }
    |> get_format()
  end

  @spec decode(ClickHouse.Result.t()) ::
          {:ok, ClickHouse.Result.t()} | {:error, ClickHouse.FormatError.t()}
  def decode(%{compressed: true} = result), do: {:ok, result}
  def decode(%{format: nil} = result), do: {:ok, result}
  def decode(result), do: result.format.decode(result)

  ################################
  # Public API
  ################################

  defp get_format(result) do
    case Config.format(result.client, result.raw_format) do
      {:ok, format} -> %{result | format: format}
      _ -> result
    end
  end
end
